"""Utilities for handling command aliases."""
from typing import List

import rich_click as click
from click import Context
from rich_click.rich_group import RichGroup


class AliasedGroup(RichGroup):
    """
    This following example implements a subclass of Group that accepts a prefix for a command.

    If there were a command called ``push``, it would accept ``pus`` as an alias (so long as it was unique)
    """

    def get_command(self, ctx: Context, cmd_name: str) -> None:
        """Given a context and a command name, this returns a Command object if it exists or returns None."""
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail(f"Too many matches: {', '.join(sorted(matches))}")

    def resolve_command(self, ctx: Context, args: List[str]) -> tuple:
        """Find the command and make sure the full command name is returned."""
        # always return the full command name
        _, cmd, args = super().resolve_command(ctx, args)
        return cmd.name, cmd, args



"""General utilities."""
import string
from collections import ChainMap
from dataclasses import asdict
from typing import TYPE_CHECKING, List, Optional

if TYPE_CHECKING:  # pragma: no-coverage
    from bumpversion.config import Config
    from bumpversion.version_part import Version


def key_val_string(d: dict) -> str:
    """Render the dictionary as a comma-delimited key=value string."""
    return ", ".join(f"{k}={v}" for k, v in sorted(d.items()))


def prefixed_environ() -> dict:
    """Return a dict of the environment with keys wrapped in `${}`."""
    import os

    return {f"${key}": value for key, value in os.environ.items()}


def labels_for_format(serialize_format: str) -> List[str]:
    """Return a list of labels for the given serialize_format."""
    return [item[1] for item in string.Formatter().parse(serialize_format) if item[1]]


def get_context(
    config: "Config", current_version: Optional["Version"] = None, new_version: Optional["Version"] = None
) -> ChainMap:
    """Return the context for rendering messages and tags."""
    import datetime

    ctx = ChainMap(
        {"current_version": config.current_version},
        {"now": datetime.datetime.now(), "utcnow": datetime.datetime.utcnow()},
        prefixed_environ(),
        asdict(config.scm_info),
        {c: c for c in ("#", ";")},
    )
    if current_version:
        ctx.new_child({f"current_{part}": current_version[part].value for part in current_version})
    if new_version:
        ctx.new_child({f"new_{part}": new_version[part].value for part in new_version})
    return ctx


def get_overrides(**kwargs) -> dict:
    """Return a dictionary containing only the overridden key-values."""
    return {key: val for key, val in kwargs.items() if val is not None}



"""Version changing methods."""
import logging
import shlex
from pathlib import Path
from typing import TYPE_CHECKING, ChainMap, List, Optional

if TYPE_CHECKING:  # pragma: no-coverage
    from bumpversion.files import ConfiguredFile
    from bumpversion.version_part import Version

from bumpversion.config import Config, update_config_file
from bumpversion.exceptions import ConfigurationError
from bumpversion.utils import get_context, key_val_string

logger = logging.getLogger("bumpversion")


def get_next_version(
    current_version: "Version", config: Config, version_part: Optional[str], new_version: Optional[str]
) -> "Version":
    """
    Bump the version_part to the next value.

    Args:
        current_version: The current version
        config: The current configuration
        version_part: Optional part of the version to bump
        new_version: Optional specific version to bump to

    Returns:
        The new version

    Raises:
        ConfigurationError: If it can't generate the next version.
    """
    if new_version:
        next_version = config.version_config.parse(new_version)
    elif version_part:
        logger.info("Attempting to increment part '%s'", version_part)
        next_version = current_version.bump(version_part, config.version_config.order)
    else:
        raise ConfigurationError("Unable to get the next version.")

    logger.info("Values are now: %s", key_val_string(next_version.values))
    return next_version


def do_bump(
    version_part: Optional[str],
    new_version: Optional[str],
    config: Config,
    config_file: Optional[Path] = None,
    dry_run: bool = False,
) -> None:
    """
    Bump the version_part to the next value or set the version to new_version.

    Args:
        version_part: The version part to bump
        new_version: The explicit version to set
        config: The configuration to use
        config_file: The configuration file to update
        dry_run: True if the operation should be a dry run
    """
    from bumpversion.files import modify_files, resolve_file_config

    ctx = get_context(config)
    version = config.version_config.parse(config.current_version)
    next_version = get_next_version(version, config, version_part, new_version)
    next_version_str = config.version_config.serialize(next_version, ctx)
    logger.info("New version will be '%s'", next_version_str)

    if config.current_version == next_version_str:
        logger.info("Version is already '%s'", next_version_str)
        return

    if dry_run:
        logger.info("Dry run active, won't touch any files.")

    ctx = get_context(config, version, next_version)

    configured_files = resolve_file_config(config.files, config.version_config)
    modify_files(configured_files, version, next_version, ctx, dry_run)

    update_config_file(config_file, config.current_version, next_version_str, dry_run)

    commit_and_tag(config, config_file, configured_files, ctx, dry_run)


def commit_and_tag(
    config: Config,
    config_file: Optional[Path],
    configured_files: List["ConfiguredFile"],
    ctx: ChainMap,
    dry_run: bool = False,
) -> None:
    """
    Commit and tag the changes, if a tool is configured.

    Args:
        config: The configuration
        config_file: The configuration file to include in the commit, if it exists
        configured_files: A list of files to commit
        ctx: The context used to render the tag and tag message
        dry_run: True if the operation should be a dry run
    """
    if not config.scm_info.tool:
        return

    extra_args = shlex.split(config.commit_args) if config.commit_args else []

    commit_files = {f.path for f in configured_files}
    if config_file:
        commit_files |= {str(config_file)}

    config.scm_info.tool.commit_to_scm(list(commit_files), config, ctx, extra_args, dry_run)
    config.scm_info.tool.tag_in_scm(config, ctx, dry_run)



"""Logging configuration for bumpversion."""
import logging

import click
from rich.logging import RichHandler

logger = logging.getLogger("bumpversion")

VERBOSITY = {
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.DEBUG,
}


def setup_logging(verbose: int = 0) -> None:
    """Configure the logging."""
    logging.basicConfig(
        level=VERBOSITY.get(verbose, logging.DEBUG),
        format="%(message)s",
        datefmt="[%X]",
        handlers=[
            RichHandler(
                rich_tracebacks=True, show_level=False, show_path=False, show_time=False, tracebacks_suppress=[click]
            )
        ],
    )
    root_logger = logging.getLogger("")
    root_logger.setLevel(VERBOSITY.get(verbose, logging.DEBUG))



"""Methods for changing files."""
import glob
import logging
from difflib import context_diff
from typing import List, MutableMapping

from bumpversion.config import FileConfig
from bumpversion.exceptions import VersionNotFoundError
from bumpversion.version_part import Version, VersionConfig

logger = logging.getLogger(__name__)


class ConfiguredFile:
    """A file to modify in a configured way."""

    def __init__(self, file_cfg: FileConfig, version_config: VersionConfig) -> None:
        self.path = file_cfg.filename
        self.parse = file_cfg.parse or version_config.parse_regex.pattern
        self.serialize = file_cfg.serialize or version_config.serialize_formats
        self.search = file_cfg.search or version_config.search
        self.replace = file_cfg.replace or version_config.replace
        self.version_config = VersionConfig(
            self.parse, self.serialize, self.search, self.replace, version_config.part_configs
        )

    def contains_version(self, version: Version, context: MutableMapping) -> bool:
        """
        Check whether the version is present in the file.

        Args:
            version: The version to check
            context: The context to use

        Raises:
            VersionNotFoundError: if the version number isn't present in this file.

        Returns:
            True if the version number is in fact present.
        """
        search_expression = self.search.format(**context)

        if self.contains(search_expression):
            return True

        # the `search` pattern did not match, but the original supplied
        # version number (representing the same version part values) might
        # match instead.

        # check whether `search` isn't customized, i.e. should match only
        # very specific parts of the file
        search_pattern_is_default = self.search == self.version_config.search

        if search_pattern_is_default and self.contains(version.original):
            # original version is present, and we're not looking for something
            # more specific -> this is accepted as a match
            return True

        # version not found
        raise VersionNotFoundError(f"Did not find '{search_expression}' in file: '{self.path}'")

    def contains(self, search: str) -> bool:
        """Does the work of the contains_version method."""
        if not search:
            return False

        with open(self.path, "rt", encoding="utf-8") as f:
            search_lines = search.splitlines()
            lookbehind = []

            for lineno, line in enumerate(f.readlines()):
                lookbehind.append(line.rstrip("\n"))

                if len(lookbehind) > len(search_lines):
                    lookbehind = lookbehind[1:]

                if (
                    search_lines[0] in lookbehind[0]
                    and search_lines[-1] in lookbehind[-1]
                    and search_lines[1:-1] == lookbehind[1:-1]
                ):
                    logger.info(
                        "Found '%s' in %s at line %s: %s",
                        search,
                        self.path,
                        lineno - (len(lookbehind) - 1),
                        line.rstrip(),
                    )
                    return True
        return False

    def replace_version(
        self, current_version: Version, new_version: Version, context: MutableMapping, dry_run: bool = False
    ) -> None:
        """Replace the current version with the new version."""
        with open(self.path, "rt", encoding="utf-8") as f:
            file_content_before = f.read()
            file_new_lines = f.newlines[0] if isinstance(f.newlines, tuple) else f.newlines

        context["current_version"] = self.version_config.serialize(current_version, context)
        context["new_version"] = self.version_config.serialize(new_version, context)

        search_for = self.version_config.search.format(**context)
        replace_with = self.version_config.replace.format(**context)

        file_content_after = file_content_before.replace(search_for, replace_with)

        if file_content_before == file_content_after and current_version.original:
            search_for_original_formatted = self.version_config.search.format(current_version=current_version.original)
            file_content_after = file_content_before.replace(search_for_original_formatted, replace_with)

        if file_content_before != file_content_after:
            logger.info("%s file %s:", "Would change" if dry_run else "Changing", self.path)
            logger.info(
                "\n".join(
                    list(
                        context_diff(
                            file_content_before.splitlines(),
                            file_content_after.splitlines(),
                            fromfile=f"before {self.path}",
                            tofile=f"after {self.path}",
                            lineterm="",
                        )
                    )
                )
            )
        else:
            logger.info("%s file %s", "Would not change" if dry_run else "Not changing", self.path)

        if not dry_run:
            with open(self.path, "wt", encoding="utf-8", newline=file_new_lines) as f:
                f.write(file_content_after)

    def __str__(self) -> str:
        return self.path

    def __repr__(self) -> str:
        return f"<bumpversion.ConfiguredFile:{self.path}>"


def resolve_file_config(files: List[FileConfig], version_config: VersionConfig) -> List[ConfiguredFile]:
    """
    Resolve the files, searching and replacing values according to the FileConfig.

    Args:
        files: A list of file configurations
        version_config: How the version should be changed

    Returns:
        A list of ConfiguredFiles
    """
    configured_files = []
    for file_cfg in files:
        if file_cfg.glob:
            configured_files.extend(get_glob_files(file_cfg, version_config))
        else:
            configured_files.append(ConfiguredFile(file_cfg, version_config))

    return configured_files


def modify_files(
    files: List[ConfiguredFile],
    current_version: Version,
    new_version: Version,
    context: MutableMapping,
    dry_run: bool = False,
) -> None:
    """
    Modify the files, searching and replacing values according to the FileConfig.

    Args:
        files: The list of configured files
        current_version: The current version
        new_version: The next version
        context: The context used for rendering the version
        dry_run: True if this should be a report-only job
    """
    _check_files_contain_version(files, current_version, context)
    for f in files:
        f.replace_version(current_version, new_version, context, dry_run)


def get_glob_files(file_cfg: FileConfig, version_config: VersionConfig) -> List[ConfiguredFile]:
    """
    Return a list of files that match the glob pattern.

    Args:
        file_cfg: The file configuration containing the glob pattern
        version_config: The version configuration

    Returns:
        A list of resolved files according to the pattern.
    """
    files = []
    for filename_glob in glob.glob(file_cfg.glob, recursive=True):
        new_file_cfg = file_cfg.copy()
        new_file_cfg.filename = filename_glob
        files.append(ConfiguredFile(new_file_cfg, version_config))
    return files


def _check_files_contain_version(
    files: List[ConfiguredFile], current_version: Version, context: MutableMapping
) -> None:
    """Make sure files exist and contain version string."""
    logger.info(
        "Asserting files %s contain the version string...",
        ", ".join({str(f.path) for f in files}),
    )
    for f in files:
        f.contains_version(current_version, context)



"""Tests for `bumpversion` package."""
import shutil
import subprocess
from pathlib import Path

import pytest
from click.testing import CliRunner, Result
from pytest import param

from bumpversion import __version__, cli
from tests.conftest import inside_dir


def test_version_displays_library_version():
    """
    Arrange/Act: Run the `version` subcommand.

    Assert: The output matches the library version.
    """
    runner: CliRunner = CliRunner()
    result: Result = runner.invoke(cli.cli, ["--version"])
    assert __version__ in result.output.strip(), "Version number should match library version."


def test_bump_no_configured_files(mocker, tmp_path):
    """
    Arrange/Act: Run the `bump` subcommand with --no-configured-files.

    Assert: There is no configured files specified to modify
    """
    mocked_do_bump = mocker.patch("bumpversion.cli.do_bump")
    runner: CliRunner = CliRunner()
    with inside_dir(tmp_path):
        result: Result = runner.invoke(cli.cli, ["--current-version", "1.0.0", "--no-configured-files", "patch"])

    if result.exit_code != 0:
        print(result.output)

    assert result.exit_code == 0

    call_args = mocked_do_bump.call_args[0]
    assert len(call_args[2].files) == 0


def test_no_configured_files_still_file_args_work(mocker, tmp_path):
    mocked_do_bump = mocker.patch("bumpversion.cli.do_bump")
    runner: CliRunner = CliRunner()
    with inside_dir(tmp_path):
        result: Result = runner.invoke(
            cli.cli, ["--current-version", "1.0.0", "--no-configured-files", "patch", "do-this-file.txt"]
        )

    if result.exit_code != 0:
        print(result.output)

    assert result.exit_code == 0

    call_args = mocked_do_bump.call_args[0]
    assert len(call_args[2].files) == 1
    assert call_args[2].files[0].filename == "do-this-file.txt"


def test_missing_explicit_config_file(tmp_path: Path):
    """The command-line processor should raise an exception if the config file is missing."""
    with inside_dir(tmp_path):
        runner: CliRunner = CliRunner()
        with inside_dir(tmp_path):
            result: Result = runner.invoke(cli.cli, ["--config-file", "missing-file.cfg"])
        assert result.exit_code != 0
        assert "'missing-file.cfg' does not exist." in result.output


def test_cli_options_override_config(tmp_path: Path, fixtures_path: Path, mocker):
    """The command-line processor should override the config file."""
    # Arrange
    config_path = tmp_path / "this_config.toml"
    fixture_toml = fixtures_path / "basic_cfg.toml"
    shutil.copy(fixture_toml, config_path)
    runner: CliRunner = CliRunner()
    mocked_do_bump = mocker.patch("bumpversion.cli.do_bump")

    # Act
    with inside_dir(tmp_path):
        result: Result = runner.invoke(
            cli.cli,
            [
                "--config-file",
                str(config_path),
                "--current-version",
                "1.1.0",
                "--new-version",
                "1.2.0",
                "--allow-dirty",
                "--parse",
                r"XXX(?P<spam>\d+);(?P<blob>\d+);(?P<slurp>\d+)",
                "--serialize",
                "XXX{spam};{blob};{slurp}",
                "--search",
                "my-search",
                "--replace",
                "my-replace",
                "--no-commit",
                "--no-tag",
                "patch",
                "do-this-file.txt",
            ],
        )

    # Assert
    assert result.exit_code == 0
    assert mocked_do_bump.call_count == 1
    assert mocked_do_bump.call_args[0][0] == "patch"
    assert mocked_do_bump.call_args[0][1] == "1.2.0"
    the_config = mocked_do_bump.call_args[0][2]
    assert the_config.current_version == "1.1.0"
    assert the_config.allow_dirty
    assert the_config.parse == r"XXX(?P<spam>\d+);(?P<blob>\d+);(?P<slurp>\d+)"
    assert the_config.serialize == ["XXX{spam};{blob};{slurp}"]
    assert the_config.search == "my-search"
    assert the_config.replace == "my-replace"
    assert the_config.commit is False
    assert the_config.tag is False
    assert len(the_config.files) == 4
    assert {f.filename for f in the_config.files} == {
        "setup.py",
        "bumpversion/__init__.py",
        "CHANGELOG.md",
        "do-this-file.txt",
    }
    assert mocked_do_bump.call_args[0][3] == config_path


@pytest.mark.parametrize(
    ["repo", "scm_command"],
    [
        param("git_repo", "git", id="git"),
        param("hg_repo", "hg", id="hg"),
    ],
)
def test_dirty_work_dir_raises_error(repo: str, scm_command: str, request):
    repo_path: Path = request.getfixturevalue(repo)
    with inside_dir(repo_path):
        # Arrange
        repo_path.joinpath("dirty2").write_text("i'm dirty! 1.1.1")
        subprocess.run([scm_command, "add", "dirty2"], check=True)
        runner: CliRunner = CliRunner()

        # Act
        result: Result = runner.invoke(cli.cli, ["patch", "--current-version", "1.1.1", "--no-allow-dirty", "dirty2"])

    # Assert
    assert result.exit_code != 0
    assert "working directory is not clean" in result.output


def test_listing_outputs_correctly_and_stops(tmp_path: Path, fixtures_path: Path):
    """The --list option should list the configuration and nothing else."""
    # Arrange
    config_path = tmp_path / "pyproject.toml"
    toml_path = fixtures_path / "basic_cfg.toml"
    shutil.copy(toml_path, config_path)
    runner: CliRunner = CliRunner()

    with inside_dir(tmp_path):
        result: Result = runner.invoke(cli.cli, ["--list", "patch"])

    if result.exit_code != 0:
        print(result.output)
        print(result.exception)

    assert result.exit_code == 0
    assert set(result.output.splitlines(keepends=False)) == {
        "new_version=1.0.1-dev",
        "current_version=1.0.0",
        "parse=(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)(\\-(?P<release>[a-z]+))?",
        "serialize=['{major}.{minor}.{patch}-{release}', '{major}.{minor}.{patch}']",
        "search={current_version}",
        "replace={new_version}",
        "tag=True",
        "sign_tags=False",
        "tag_name=v{new_version}",
        "tag_message=Bump version: {current_version} → {new_version}",
        "allow_dirty=False",
        "commit=True",
        "message=Bump version: {current_version} → {new_version}",
        "commit_args=None",
        "files=[{'filename': 'setup.py', 'glob': None, 'parse': None, 'serialize': "
        "None, 'search': None, 'replace': None}, {'filename': "
        "'bumpversion/__init__.py', 'glob': None, 'parse': None, 'serialize': None, "
        "'search': None, 'replace': None}, {'filename': 'CHANGELOG.md', 'glob': None, "
        "'parse': None, 'serialize': None, 'search': '**unreleased**', 'replace': "
        "'**unreleased**\\n**v{new_version}**'}]",
    }


def test_non_scm_operations_if_scm_not_installed(tmp_path: Path, monkeypatch):
    """Everything works except SCM commands if the SCM is unusable."""
    # Arrange
    monkeypatch.setenv("PATH", "")

    with inside_dir(tmp_path):
        version_path = tmp_path / "VERSION"
        version_path.write_text("31.0.3")

        runner: CliRunner = CliRunner()

        # Act
        runner.invoke(cli.cli, ["major", "--current-version", "31.0.3", "VERSION"])

    # Assert
    assert version_path.read_text() == "32.0.0"



"""The setup script."""
from pathlib import Path
from typing import List, Optional

from setuptools import setup


def parse_reqs_in(filepath: Path, visited: Optional[set] = None) -> List[str]:  # noqa: C901
    """
    Parse a file path containing a pip-tools requirements.in and return a list of requirements.

    Will properly follow ``-r`` and ``-c`` links like ``pip-tools``. This
    means layered requirements will be returned as one list.

    Other ``pip-tools`` and ``pip``-specific lines are excluded.

    Args:
        filepath (Path): The path to the requirements file
        visited (set, optional): A set of paths that have already been visited.

    Returns:
        All the requirements as a list.
    """
    if visited is None:
        visited = set()
    reqstr: str = filepath.read_text()
    reqs: List[str] = []
    for line in reqstr.splitlines(keepends=False):
        line = line.strip()  # noqa: PLW2901
        if not line:
            continue
        elif not line or line.startswith("#"):
            # comments are lines that start with # only
            continue
        elif line.startswith("-c"):
            _, new_filename = line.split()
            new_file_path = filepath.parent / new_filename.replace(".txt", ".in")
            if new_file_path not in visited:
                visited.add(new_file_path)
                reqs.extend(parse_reqs_in(new_file_path, visited))
        elif line.startswith(("-r", "--requirement")):
            _, new_filename = line.split()
            new_file_path = filepath.parent / new_filename
            if new_file_path not in visited:
                visited.add(new_file_path)
                reqs.extend(parse_reqs_in(new_file_path, visited))
        elif line.startswith("-f") or line.startswith("-i") or line.startswith("--"):
            continue
        elif line.startswith("-Z") or line.startswith("--always-unzip"):
            continue
        else:
            reqs.append(line)
    return reqs


here: Path = Path(__file__).parent.absolute()
requirements = parse_reqs_in(here / "requirements/prod.in")
dev_requirements = parse_reqs_in(here / "requirements/dev.in")
test_requirements = parse_reqs_in(here / "requirements/test.in")

setup(
    install_requires=requirements,
    tests_require=test_requirements,
    extras_require={"dev": dev_requirements, "test": test_requirements},
)
